'use strict';

var OmniTypes = require('../types');
var OmniSchema = require('../schemas');
var _ = require('lodash');

// db-mongoosejs.js
// An OmniSchema plugin to generate a Mongoose schema, enabling ODM for MongoDB
// Uses: Mongoose: http://mongoosejs.com/

var mongoose = require('mongoose');

// Hints for effective plugin development:
// 1. Store all plugin specific properties inside a name spaced property
//
// 2. Be sure to define plugin properties for AT LEAST the data types 
//    String, Number, and Boolean (they are the Javascript primatives).
//
// 3. If multiple onType functionality specifications are used, be sure
//    to list them in the order of most general to most specific
//
// 4. Try to define onType functionality that matches 'enumValues'
//	  to ensure there is default behavior for any enumerations.
//
// 5. Create a namespace off of the OmniSchema type to prevent name clashes
//    from other plugins.

// Our namespace for this plugin
OmniSchema.mongoose = {};
OmniSchema.mongoose.schemas = {};
OmniSchema.mongoose.models = {};

// Our "define..Type" method to decorate the data types with attributes that
// tell us which Mongoose schema types to use for each type...
OmniSchema.mongoose.defineDbType = function (name, mongooseType, props) {
	var mongoosedb = { mongoosedb: Object.assign({ type: mongooseType }, props) };
	OmniTypes.defineDataType(name, mongoosedb);
};

var plugin = function plugin() {

	var defineDbType = OmniSchema.mongoose.defineDbType;

	// MUST at least define entries for Javascript primatives...
	defineDbType('String', String, { minValProperty: 'minlength', maxValProperty: 'maxlength' });
	defineDbType('Number', Number, { minValProperty: 'min', maxValProperty: 'max' });
	defineDbType('Boolean', Boolean);
	defineDbType('Object', Object);
	defineDbType('DateTime', Date);

	// And now, define specialized controls for the specialzed data types that vary
	// from our defaults. Inheritance will take care of the rest...
	defineDbType('Email', String, { match: /^[\w-\.]+@([\w-]+\.)+[\w-]{2,4}$/ });

	// Now, mix in functionality to OmniSchema...
	OmniSchema.mixin({

		onSchema: [{ func: function getMongooseSchema() {

				// If we have already defined this schema, return the singleton...
				if (this.collectionName && OmniSchema.mongoose.schemas[this.collectionName]) {
					return OmniSchema.mongoose.schemas[this.collectionName];
				}

				var omniSchema = this;
				var mongooseSchemaDef = {};
				var fl = this.getFieldList();
				var hasVirtualFields = false;
				var _iteratorNormalCompletion = true;
				var _didIteratorError = false;
				var _iteratorError = undefined;

				try {
					for (var _iterator = fl[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
						var _fieldName = _step.value;

						if (_fieldName !== '_id') {
							var _field = omniSchema.getField(_fieldName);
							if (_field instanceof OmniSchema.OmniField) {
								var mongooseField = _field.getMongooseField();
								if (mongooseField) {
									mongooseSchemaDef[_fieldName] = mongooseField;
								}
							}

							if (_field.type instanceof OmniSchema && _field.foreignKeyField) {
								hasVirtualFields = true;
							}
						}
					}
				} catch (err) {
					_didIteratorError = true;
					_iteratorError = err;
				} finally {
					try {
						if (!_iteratorNormalCompletion && _iterator.return) {
							_iterator.return();
						}
					} finally {
						if (_didIteratorError) {
							throw _iteratorError;
						}
					}
				}

				var options = {};

				if (omniSchema.hasChildren || omniSchema.isSubClass) {
					// This schema is part of a hierarchy...
					options.discriminatorKey = '_type_';
				}

				if (omniSchema.isSubClass) {
					options._id = false;
				}

				if (hasVirtualFields) {
					options.toJSON = { virtuals: true };
				}

				var finalSchema = new mongoose.Schema(mongooseSchemaDef, options);

				// If there are schema references with a foreign key, those
				// need to be optimized as "virtual" fields...
				if (hasVirtualFields) {
					var _iteratorNormalCompletion2 = true;
					var _didIteratorError2 = false;
					var _iteratorError2 = undefined;

					try {
						for (var _iterator2 = fl[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
							var fieldName = _step2.value;

							var field = omniSchema.getField(fieldName);
							if (field.type instanceof OmniSchema && field.foreignKeyField) {
								// Define a "virtual" field for this one...
								finalSchema.virtual(fieldName, { ref: field.type.collectionName, localField: '_id', foreignField: field.foreignKeyField });
							}
						} // for
					} catch (err) {
						_didIteratorError2 = true;
						_iteratorError2 = err;
					} finally {
						try {
							if (!_iteratorNormalCompletion2 && _iterator2.return) {
								_iterator2.return();
							}
						} finally {
							if (_didIteratorError2) {
								throw _iteratorError2;
							}
						}
					}
				}

				if (this.collectionName) {
					OmniSchema.mongoose.schemas[this.collectionName] = finalSchema;
				}

				return finalSchema;
			}
		}, { func: function getMongooseModel() {

				var className = this.collectionName;

				if (className && OmniSchema.mongoose.models[className]) {
					return OmniSchema.mongoose.models[className];
				}

				var model = void 0;

				if (this.isSubClass) {
					var rootModel = this.rootSchema.getMongooseModel();
					if (!rootModel) {
						throw new Error('Can not create Mongoose model for ' + className + ': the root mongoose schema for ' + this.rootSchema.collectionName + ' has not been defined yet.');
					}
					model = rootModel.discriminator(className, this.getMongooseSchema());
				} else {
					model = mongoose.model(className, this.getMongooseSchema());
				}

				OmniSchema.mongoose.models[className] = model;

				return model;
			}

		}],

		onField: {
			func: function getMongooseField() {

				if (this.type instanceof OmniSchema) {
					// We have a reference to another schema...
					if (this.persistence === 'embed') {
						// Embedded schemas are defined directly in this schema...
						var mschema = this.type.getMongooseSchema();
						if (this.isArray) {
							return [mschema];
						} else {
							return mschema;
						}
					} else {

						if (this.foreignKeyField) {
							// Fields that have foreign keys are optimized as virtual fields.
							// see getMongooseSchema()
							return undefined;
						}

						var ref = { ref: this.type.collectionName, type: mongoose.Schema.Types.ObjectId };
						if (this.isArray) {
							return [ref];
						} else {
							return ref;
						}
					}
				}

				if (!this.type.getMongooseField) {
					throw new Error('getMongooseField has not been defined for datatype ' + this.type.name);
				}
				return this.type.getMongooseField(this);
			}
		},

		// For Types, always define in the order of most general to most specific...
		onType: [{ matches: 'mongoosedb',

			func: function getMongooseField(field, controlProps) {

				var mongooseField = {
					type: this.mongoosedb.type
				};

				if (field.isRequired) {
					mongooseField.required = true;
				}

				if (_.has(field, 'validation.min') && this.mongoosedb.minValProperty) {
					mongooseField[this.mongoosedb.minValProperty] = field.validation.min;
				}

				if (_.has(field, 'validation.max') && this.mongoosedb.maxValProperty) {
					mongooseField[this.mongoosedb.maxValProperty] = field.validation.max;
				}

				if (_.has(field, 'db.unique')) {
					mongooseField.unique = field.db.unique;
				}

				if (this.mongoosedb.type === String && this.mongoosedb.match) {
					mongooseField.match = this.mongoosedb.match;
				}

				if (this.mongoosedb.type === String && this.enumValues) {
					mongooseField.enum = this.enumValues.map(function (ev) {
						return ev.value;
					});
				}

				if (field.isArray) {
					mongooseField.type = [mongooseField.type];
				}

				return mongooseField;
			}

		}]

	});
};

module.exports = { plugin: plugin };