'use strict';

var OmniTypes = require('../types');
var OmniSchema = require('../schemas');
var _ = require('lodash');

// mock-faker.js
// An OmniSchema plugin to generate mock data for a record that matches the schema.
// Uses: Faker: https://github.com/stympy/faker

var faker = require('faker');

// Hints for effective plugin development:
// 1. Store all plugin specific properties inside a name spaced property
//
// 2. Be sure to define plugin properties for AT LEAST the data types 
//    String, Number, and Boolean (they are the Javascript primatives).
//
// 3. If multiple onType functionality specifications are used, be sure
//    to list them in the order of most general to most specific
//
// 4. Try to define onType functionality that matches 'enumValues'
//	  to ensure there is default behavior for any enumerations.
//
// 5. Create a namespace off of the OmniSchema type to prevent name clashes
//    from other plugins.

// Our namespace for this plugin
OmniSchema.faker = {};

// Our "define..Type" method to decorate the data types with attributes that
// tell us which Faker functions to use for each type...
OmniSchema.faker.defineFakerType = function (name, fakerFunc) {
	var fakerSpec = { fakerSpec: { getMockData: fakerFunc } };
	OmniTypes.defineDataType(name, fakerSpec);
};

var plugin = function plugin() {

	var defineFakerType = OmniSchema.faker.defineFakerType;

	function fakeCurrency() {
		return Number(faker.finance.amount());
	}

	// MUST at least define entries for Javascript primatives...
	defineFakerType('String', function (field) {
		return faker.lorem.words();
	});

	defineFakerType('Number', function (field) {
		var params = { min: -999, max: 999 };
		if (typeof field.minVal !== 'undefined') {
			params.min = field.minVal;
		}
		if (typeof field.maxVal !== 'undefined') {
			params.max = field.maxVal;
		}
		return faker.random.number(params);
	});

	defineFakerType('Boolean', function (field) {
		return faker.random.boolean();
	});
	defineFakerType('DateTime', function (field) {
		return faker.date.recent();
	});

	// And now, define specialized controls for the specialzed data types that vary
	// from our defaults. Inheritance will take care of the rest...
	defineFakerType('Text', function (field) {
		return faker.lorem.paragraph();
	});
	defineFakerType('FullName', function () {
		return faker.name.firstName() + ' ' + faker.name.lastName();
	});
	defineFakerType('FirstName', function (field) {
		return faker.name.firstName();
	});
	defineFakerType('LastName', function (field) {
		return faker.name.lastName();
	});
	defineFakerType('Password', function (field) {
		return faker.internet.password();
	});
	defineFakerType('Phone', function (field) {
		return faker.phone.phoneNumber();
	});
	defineFakerType('Email', function (field) {
		return faker.internet.email();
	});
	defineFakerType('Url', function (field) {
		return faker.internet.url();
	});
	defineFakerType('StreetAddress', function (field) {
		return faker.address.streetAddress();
	});
	defineFakerType('City', function (field) {
		return faker.address.city();
	});
	defineFakerType('State', function (field) {
		return faker.address.stateAbbr();
	});
	defineFakerType('PostalCode', function (field) {
		return faker.address.zipCode();
	});
	defineFakerType('CreditCardNumber', function () {
		return '4116817818840415';
	});

	defineFakerType('Integer', function (field) {
		var params = { min: -999, max: 999 };
		if (typeof field.minVal !== 'undefined') {
			params.min = field.minVal;
		}
		if (typeof field.maxVal !== 'undefined') {
			params.max = field.maxVal;
		}
		return Math.trunc(faker.random.number(params));
	});

	defineFakerType('Currency', function (field) {
		return fakeCurrency();
	});

	defineFakerType('Sex', function (field) {
		return faker.random.boolean() ? 'F' : 'M';
	});

	// Now, mix in functionality to OmniSchema...
	OmniSchema.mixin({

		onSchema: [{
			func: function getMockData(uiExclude) {

				if (this.__mockRecursivePrevent) {
					return;
				} else {
					this.__mockRecursivePrevent = true;
				}

				var omniSchema = this;
				var mockData = {};
				var fl = this.getFieldList();
				var _iteratorNormalCompletion = true;
				var _didIteratorError = false;
				var _iteratorError = undefined;

				try {
					for (var _iterator = fl[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
						var fieldName = _step.value;

						var field = omniSchema.getField(fieldName);
						if (field instanceof OmniSchema.OmniField && (!uiExclude || !field.uiExclude)) {
							if (field.isArray) {
								var mockArray = [];
								for (var i = 0; i < 3; i++) {
									if (Math.random() > 0.5) {
										mockArray.push(field.getMockData());
									}
								}
								mockData[fieldName] = mockArray;
							} else {
								mockData[fieldName] = field.getMockData(uiExclude);
							}
						}
					}
				} catch (err) {
					_didIteratorError = true;
					_iteratorError = err;
				} finally {
					try {
						if (!_iteratorNormalCompletion && _iterator.return) {
							_iterator.return();
						}
					} finally {
						if (_didIteratorError) {
							throw _iteratorError;
						}
					}
				}

				this.__mockRecursivePrevent = false;
				return mockData;
			}
		}],

		onField: {
			func: function getMockData(uiExclude) {
				if (!this.type.getMockData) {
					throw new Error('getMockData has not been defined for datatype ' + this.type.name);
				}
				if (this.type instanceof OmniSchema) {
					return this.type.getMockData(uiExclude);
				} else {
					return this.type.getMockData(this);
				}
			}
		},

		// For Types, always define in the order of most general to most specific...
		onType: [{ matches: 'fakerSpec',

			func: function getMockData(field) {
				return this.fakerSpec.getMockData(field);
			}

		}]

	});
};

module.exports = { plugin: plugin };